#version 330 core

const int maxNumberOfLights = 4;

const vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

struct LightSource {
	vec3 position; //Position in worldspace
	vec3 color;
	vec3 spotDirection;
	//attenuation
	float constantAttenuation;
	float linearAttenuation;
	float quadraticAttenuation;

	int on; //0 if off, 1 if on
	int isSpot; //1 if spotlight
	float spotCutoff; 
	float spotExponent;
};

struct Material {
	vec3 diffuse;
	vec3 ambient;
	vec3 specular;
	float shininess;
};

uniform float xOffset;
uniform float yOffset;

in vec2 fragmentUV;
in vec3 position_worldspace;
in vec3 position_cameraspace;
in vec4 shadowCoordinate;

in vec3 Normal_cameraspace;
in vec3 LightDirection_tangentspace[maxNumberOfLights];
in vec3 EyeDirection_tangentspace;

out vec3 fragColor;

uniform sampler2D colorTexture;
uniform sampler2D NormalMapTex;
//uniform sampler2D shadow_map;
uniform sampler2DShadow shadow_map;

uniform mat4 modelViewMatrix;
uniform int numberOfLights;
uniform LightSource lights[maxNumberOfLights];
uniform Material material;

Material mat;

float random(vec3 seed, int i){
	vec4 seed4 = vec4(seed,i);
	float dot_product = dot(seed4, vec4(12.9898,78.233,45.164,94.673));
	return fract(sin(dot_product) * 43758.5453);
}

vec3 ads( int lightIndex, vec3 normal )
{
	vec3 lightVec = (lights[lightIndex].position - position_worldspace);
	float d = length(lightVec); // distance to light
	

	//calculate Ambient Term:  
	vec3 Iamb = 0.1+mat.ambient;

	//Direction of the light (from the fragment to the light
	vec3 l = normalize(LightDirection_tangentspace[lightIndex]);

	float sDotN = clamp( dot(l,normal), 0 , 1);
	vec3 R = reflect( -l, normal );
	//calculate Diffuse Term:
	vec3 Idiff = mat.diffuse * sDotN;
	Idiff = clamp(Idiff, 0.0, 1.0);

	vec3 E = normalize(EyeDirection_tangentspace);

	// calculate Specular Term:
	vec3 Ispec = vec3(0.0);
	if( sDotN > 0.0 ){
		vec3 Ispec = mat.specular * pow(clamp(dot(E,R),0,1),mat.shininess);
		Ispec = clamp(Ispec, 0.0, 1.0); 
	}

	//For ammount of light
	float visibility = 1.0;
	float spotEffect = 1.0;

	if (lights[lightIndex].isSpot == 1) {

		float intensity = 0.0;

		vec3 ld = normalize(lightVec);  
		vec3 sd = normalize(lights[lightIndex].spotDirection);
		spotEffect = dot(sd,-ld);
		if (spotEffect < cos(lights[lightIndex].spotCutoff)) 
		{
			spotEffect = smoothstep(lights[lightIndex].spotCutoff-0.02,     
									lights[lightIndex].spotCutoff, spotEffect);
		} else spotEffect = 1.0;
		// inside the cone?
		if (spotEffect > cos(lights[lightIndex].spotCutoff)) {
			float bias = 0.01;
			spotEffect = pow(spotEffect, lights[lightIndex].spotExponent);
			//If is in cone then full intensity
			intensity = 1.0;
			// Avoid counter shadow
			for (int i =0 ; i < 16  ; i++){
				int index = int(16.0*random(gl_FragCoord.xyy, i))%16;
				intensity -= (0.8/16.0)*
					(1-textureProj(
						shadow_map,
						vec4(
							shadowCoordinate.xy + 
							(poissonDisk[index]/700.0)*shadowCoordinate.w,
							shadowCoordinate.z - bias,
							shadowCoordinate.w
						)));
			}
		} else { //if not in cone then reduce to 0.1
			intensity = 0.1;
		}

		//In cone intensity * shadow map visibility
		visibility = visibility * intensity;
		
	}


	// attenuation due to distance fall off times the spotFactor
	float attenuation = 1.0 / (lights[lightIndex].constantAttenuation +
						lights[lightIndex].linearAttenuation * d +
						lights[lightIndex].quadraticAttenuation *d*d);

	  
	return lights[lightIndex].color * (Iamb + visibility * Idiff + visibility * Ispec) * attenuation;
}

void main()
{
	vec3 textureColor = texture(colorTexture, fragmentUV).rgb;
	vec3 TextureNormal_tangentspace = texture( NormalMapTex, fragmentUV).rgb*2.0 - 1.0;
	

	mat = material;

	mat.ambient *= textureColor;
	mat.diffuse *= textureColor;
	mat.specular *= textureColor;
	
	vec3 contrib = vec3(0.0);

	//For all lights
	for(int i=0; i < numberOfLights; i++) {
	//If the light is turned on
		if(lights[i].on == 1){
			contrib += ads(i,TextureNormal_tangentspace);
			//contrib += LightDirection_tangentspace[i];
		}
	}

	fragColor = contrib;
}

